/**
* @file  filedevice.cpp
* @brief Implementierung der Klasse FileDevice.
*/
//////////////////////////////////////////////////////////////////////

#include "filedevice.h"
#include "../Common/softwareinfo.h"
#include <time.h>

FileDevice::FileDevice()
{
    // Initialisieren
    reset();
    // Fehlerobjekt zuruecksetzen
    resetError();
}

FileDevice::~FileDevice()
{
}

void FileDevice::reset()
{
    // Basis Reset
    FileIntelHex::reset();

    // Dateiname der geladenen Datei halten
    m_fileName.clear();

    // Header loeschen
    memset( m_headBuf, 0, sizeof(m_headBuf) );
    m_header = (SDFFileDeviceHeader*)m_headBuf;

    // Firmwaredaten loeschen
    m_fileRawFirmware.reset();
    // Anzeigedaten loeschen
    m_fileDisplay.reset();
    // Geraetetexte loeschen
    m_fileDeviceText.reset();
}

unsigned int FileDevice::GetHeaderVersionFile()
{ 
    return m_header->versionFile;
}

unsigned int FileDevice::GetHeaderVersionTimestamp( std::string &timestamp )
{
    timestamp = boost::str( boost::format( "%1%" ) % std::string( m_header->timestamp ) );
	return timestamp.size();
}

unsigned int FileDevice::GetHeaderVersionFirmware( std::string &versionFirmware )
{
	versionFirmware = boost::str( boost::format( "%1%" ) % std::string( m_header->versionFirmware ) );
    return versionFirmware.size();
}

CDFSoftwarebuildVersion FileDevice::GetHeaderVersionFirmware()
{
    return CDFSoftwarebuildVersion( m_header->versionFirmware );
}

StructSoftwareInfo FileDevice::GetHeaderSoftwareInfo()
{
    return m_header->softwareInfo;
}

SDFFileDeviceHeader FileDevice::GetHeader( const std::string &filePath )
{
    // Lesen der Hexdatei
    FILE *fp;
    char data[0x400];

    unsigned char headerBuf[IV_DFFILEDEVICE_SIZE_HEADER];
    memset(&headerBuf, 0, sizeof(headerBuf));
    SDFFileDeviceHeader *header = (SDFFileDeviceHeader*)headerBuf;

    if ( (fp = fopen(filePath.data(), "r")) == (FILE*)0 )
    {
        // Oeffnen der Datei gescheitert
		setErrorState( FileOpen );
		setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        return *header;
    }

    unsigned int numLine = 0;
    unsigned char count;
    unsigned short i, address;

    while( !feof( fp ) && (errorState() == None) )
    {
        if ( fgets( data, 0x400, fp ) != 0 )
        {
            switch ( GetType( data ) )
            {
                case TypeDeviceHeader:

                    address = GetAddress( data );
                    count   = GetCount( data );
                    // spaetere Groessenanderung vorbehalten!
                    for ( i = address; (i < IV_DFFILEDEVICE_SIZE_HEADER) && ((i - address) < count); ++i )
                        headerBuf[i] = GetData8( data, i - address );

                    break;
            }

            numLine++;
        }
    }

    fclose( fp );	

    // Abgeaenderter Header ab 03.01.33 ist groesser 4 Byte
    if ( header->sizeData ) // Vorher war nur die Version vorhanden!
    {
        // CRC16 von Header pruefen!
        // Der crcHead ist immer der letzte 16bit Wert in der Struktur und
        // geht ueber den gesamten Dateiheader, sich selbst ausgeschlossen.
        if ( *(unsigned short*)&headerBuf[header->sizeData - sizeof(unsigned short)] != GetCRC16( headerBuf, header->sizeData - sizeof(unsigned short) ) )
        {
            // Fehler zuweisen
			setErrorState( HeadCrc );
			setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );

            memset( &headerBuf, 0, sizeof(headerBuf) );
            return *header;
        }

        // Aeltere Struktur auf aktuell verwendete Version anpassen.
        if ( header->sizeData < sizeof(SDFFileDeviceHeader) )
        {
            // CRC-Position hat eine neue Bedeutung, vorsichtshalber hier löschen.
            *(unsigned short*)&headerBuf[header->sizeData - sizeof(unsigned short)] = 0;

            //############################################
            // SoftwareInfo ggf. auf die aktuelle Version anpassen.
            header->softwareInfo = CheckSoftwareInfo( header->softwareInfo );

            // CRC nach hinten verschieben, jedoch die Struktur-Version (sizeData) beibehalten.
            header->crcHead = GetCRC16( headerBuf, header->sizeData - sizeof(unsigned short) );
        }
        else if ( header->sizeData > sizeof(SDFFileDeviceHeader) )
        {
            // Fehler zuweisen
			setErrorState( Version );
			setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );

            memset( &headerBuf, 0, sizeof(headerBuf) );
            return *header;
        }
    }
    else
    {
        // Fehler zuweisen
		setErrorState( FunctionUnavailable );
		setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );

        memset( &headerBuf, 0, sizeof(headerBuf) );
        return *header;
    }

    return *header;
}

int FileDevice::CheckVersionFile( SDFFileDeviceHeader &header )
{
    // Version auswerten (Majornumber muss uebereinstimmen) Erst ab: 03.01.33
    if ( ((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_PZE   & 0xFFFF0000))  
        && ((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_ZK    & 0xFFFF0000)) 
        && ((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_BDE   & 0xFFFF0000)) 
        && ((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_MOBIL & 0xFFFF0000)) 
        && ((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_MDE   & 0xFFFF0000)) 
        && ((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_AE        & 0xFFFF0000)) 
        && ((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_TIMEBOY   & 0xFFFF0000)) 
        && ((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_MULTIPLE  & 0xFFFF0000)) 
        && ((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_FLEX      & 0xFFFF0000)) 
        && ((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_EXKLUSIV  & 0xFFFF0000)) 
        && ((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_TIMEBOY_MOBIL_PZE & 0xFFFF0000))
        && ((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_TERMINAL_2_8 & 0xFFFF0000))
        && ((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_TERMINAL_4_3 & 0xFFFF0000))
        && ((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_IPC_EXTENSION_VARIO_5_7 & 0xFFFF0000))
        && ((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_IPC_EXTENSION_VARIO_10 & 0xFFFF0000))
        && ((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_IPC_EXTENSION_Q7 & 0xFFFF0000))
        && ((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_IO_BOX & 0xFFFF0000))
        && ((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_ZK_BOX & 0xFFFF0000))
        && ((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_MOBIL_BOX & 0xFFFF0000))
		&& ((header.versionFile & 0xFFFF0000) != (IV_DFFILEDEVICE_VERSION_SOFTWARE_INFO & 0xFFFF0000)) )
    {
        // Fehler zuweisen
		setErrorState( DeviceUnknown );
		setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        return 0;
    }

    return 1;
}

unsigned int FileDevice::GetVersionFile( const std::string &filePath )
{
    SDFFileDeviceHeader header = GetHeader( filePath );

    if ( !header.sizeData || !CheckVersionFile( header ) )
        return 0;

    return header.versionFile;
}

unsigned int FileDevice::GetVersionFirmware( const std::string &filePath, std::string &versionFirmware )
{
    SDFFileDeviceHeader header = GetHeader( filePath );

    if ( !header.sizeData || !CheckVersionFile( header ) )
        return 0;

	versionFirmware = boost::str( boost::format( "%1%" ) % std::string( header.versionFirmware ) );
    return versionFirmware.size();
}

bool FileDevice::load( const std::string &filePath ) 
{
    // Umgebungsreset
    reset();
    // Fehlerobjekt zuruecksetzen
    resetError();

    // Header lesen
    *m_header = GetHeader( filePath.data() );

    if ( !m_header->sizeData || !CheckVersionFile( *m_header ) )
    {
        // Header loeschen
        memset( m_headBuf, 0, sizeof(m_headBuf) );
        m_header = (SDFFileDeviceHeader*)m_headBuf;
        return 0;
    }

    // Dateiname halten
    m_fileName = filePath;

    // Firmwaredaten laden
    if ( !m_fileRawFirmware.load( filePath ) )
        return 0;

	unsigned int first = m_fileRawFirmware.firstAddress();
	unsigned int last = m_fileRawFirmware.lastAddress();
	unsigned int size = m_fileRawFirmware.sizeData();

    // CRC32 von Program-Flash pruefen
    if ( m_header->crcProgram != m_fileRawFirmware.checksum32bit() )
    {
        // Fehler zuweisen
		setErrorState( ProgramCrc );
		setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        return 0;
    }

    // Displaydaten laden
    if ( !m_fileDisplay.Load( filePath.data() ) )
    {
        // Keine Displaydaten vorhanden?
        if ( NoData != m_fileDisplay.errorState() )
        {
			setErrorState( LoadDisplayData );
			setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
            return 0;
        }
    }
    // Geraetetexte laden
    if ( !m_fileDeviceText.Load( filePath.data() ) )
    {
        // Keine Textdaten vorhanden?
        if ( NoData != m_fileDeviceText.errorState() )
        {
			setErrorState( LoadDeviceText );
			setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
            return 0;
        }
    }

    return 1;
}

bool FileDevice::save( const std::string &filePath, SDFFileDeviceHeader &header ) 
{
    // Fehlerobjekt zuruecksetzen
    resetError();

    //############################
    // Headerangaben pruefen

    // Eingetragene Strukturgroesse sicherstellen
    if ( !header.sizeData )
        header.sizeData = sizeof(SDFFileDeviceHeader);

    // Dateiversionsangabe
    if ( !CheckVersionFile( header ) )
        return 0;

    // Firmwareversion pruefen
    if ( (strlen( header.versionFirmware ) < 8) || (strlen( header.versionFirmware ) > 32) )
    {
        // Fehler zuweisen
		setErrorState( FirmwareVersionLength );
		setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        return 0;
    }

    // Builddatum, wenn nicht vorhanden, eintragen
    if ( strlen( header.timestamp ) != 19 )
    {
        time_t ltime;
        struct tm *ltm;

        time( &ltime );
        ltm = localtime( &ltime );
        strftime( header.timestamp, sizeof(header.timestamp), "%Y-%m-%d %H:%M:%S", ltm );
    }

    // CRC32 fuer Program-Flash berechnen
    header.crcProgram = m_fileRawFirmware.checksum32bit();

    // CRC16 fuer Dateiheader berechnen
    unsigned char *pCRC = &((unsigned char*)&header)[header.sizeData - sizeof(unsigned short)];
    *(unsigned short*)pCRC = GetCRC16( (unsigned char*)&header, header.sizeData - sizeof(unsigned short) );

    // Header loeschen
    memset( m_headBuf, 0, sizeof(m_headBuf) );
    m_header = (SDFFileDeviceHeader*)m_headBuf;
    // Struktur uebernehmen
    *m_header = header;


    //############################
    // Schreiben der Hexdatei

    FILE *fp;
    char data[0x400];
    unsigned short i;
    unsigned int numLine = 0;

    if ( (fp = fopen( filePath.data(), "w" )) == (FILE*)0 )
    {
        // Oeffnen der Datei gescheitert
		setErrorState( FileOpen );
		setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        return 0;
    }

    // Dateiheader schreiben
    for ( i = 0; i < header.sizeData; i += maxDataSize() )
    {
        // Extended Segment Address Record
        if ( !MakeTypedDataLine( i, TypeDeviceHeader, &m_headBuf[i],
                (header.sizeData - i) > maxDataSize() ? maxDataSize() : header.sizeData - i, data ) )
        {
            // Fehler zuweisen
			setErrorState( RecordMake );
			setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
            break;
        }

        fprintf( fp, "%s\n", data );
        numLine++;
    }

    // Abbruch nach Fehler
    if ( errorState() != None )
    {
        fclose( fp );
        return 0;
    }

    fclose( fp );

    // Abbruch nach Fehler
    if ( errorState() != None )
        return 0;

    // Firmwaredaten schreiben
    if ( !m_fileRawFirmware.save( filePath, true ) )
        return 0;

    // Anzeigedaten in Datei schreiben
    if ( m_fileDisplay.GetSizeData() )
    {
        if ( !m_fileDisplay.Save( filePath, true ) )
            return 0;
    }

    // Geraetetexte
    if ( m_fileDeviceText.GetSizeData() )
    {
        if ( !m_fileDeviceText.Save( filePath, true ) )
            return 0;
    }

    // Endezeile eintragen
    if ( (fp = fopen( filePath.data(), "a" )) == (FILE*)0 )
    {
        // Oeffnen der Datei gescheitert
		setErrorState( FileOpen );
		setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        return 0;
    }

    if ( !MakeEndLine( data ) )
    {
        fclose( fp );
        // Fehler zuweisen
		setErrorState( RecordEndMake );
		setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        return 0;
    }

    fprintf( fp, "%s\n", data );
    numLine++;

    fclose( fp );

    return 1;
}

int FileDevice::build( const std::string &filePath, SDFFileDeviceHeader &header, const std::string &fileFirmware, const std::string &fileDisplayData, const std::string &fileDeviceText )
{
    // Umgebungsreset
    reset();
    // Fehlerobjekt zuruecksetzen
    resetError();

    // Laden der Firmware
    if ( !m_fileRawFirmware.load( fileFirmware ) )
        return 0;

    if ( fileDisplayData.size() )
    {
        // Laden der Anzeigedaten
        if ( !m_fileDisplay.Load( fileDisplayData ) )
            return 0;
    }

    if ( fileDeviceText.size() )
    {
        // Importieren der Geraetetexte
        if ( !m_fileDeviceText.Import( fileDeviceText ) )
            return 0;
    }

    // Schreiben der Geraetedatei
    if ( !save( filePath, header ) )
        return 0;

    return 1;
}

StructSoftwareInfo FileDevice::SoftwareInfoStringToStruct( const std::string &str )
{
	std::string tmp( str );
    StructSoftwareInfo swInfo;
    memset( &swInfo, 0, sizeof(StructSoftwareInfo) );

    // # durch : ersetzen
    tmp[INTELHEX_START] = INTELHEX_START_SIGN;    // start of record
    // CRC pruefen
    if ( !CheckCrc( tmp.data() ) )
        return swInfo;

    unsigned char i, count = GetCount( tmp.data() );
    // spaetere Groessenanderung vorbehalten!
    for ( i = 0; (i < count) && i < sizeof(StructSoftwareInfo); ++i )
        ((unsigned char*)&swInfo)[i] = GetData8(tmp.data(), i);

    return CheckSoftwareInfo( swInfo );
}

bool FileDevice::SoftwareInfoStructToString( std::string &str, StructSoftwareInfo &swInfo )
{
	str.resize( INTELHEX_LENGTH_FRAME + sizeof(StructSoftwareInfo) * 2 );
    if ( MakeDataLine( 0, (unsigned char*)&swInfo, sizeof(StructSoftwareInfo), const_cast<char*>( str.data() ) ) == 0 )
        return false;

    // : durch # ersetzen
    str[INTELHEX_START] = '#';    // start of record

    return true;
}

StructSoftwareInfo FileDevice::CheckSoftwareInfo( StructSoftwareInfo swInfo )
{
    return _deviceInfo.checkInfo( swInfo );
}

unsigned int FileDevice::GetSizeData()
{
    return m_fileRawFirmware.sizeData();
}

FileDisplay *FileDevice::GetFileDisplay()
{
    return &m_fileDisplay;
}

FileDeviceText *FileDevice::GetFileDeviceText()
{
    return &m_fileDeviceText;
}

int FileDevice::CompareInfos( StructHardwareInfo hwInfo, StructSoftwareInfo swInfo )
{
    return _deviceInfo.compareInfos( hwInfo, swInfo );
}

int FileDevice::SoftwareInfoToString( StructSoftwareInfo swInfo, std::string &info )
{
    if ( !_deviceInfo.SoftwareInfoToString( swInfo, info ) )
    {
        setErrorText( _deviceInfo.m_errorText );
        return false;
    }

    return true;
}

